@@ -8,24 +8,25 @@ import functools
 import itertools
 import uuid
 from typing import (
+    TYPE_CHECKING,
     Any,
     AsyncIterable,
+    Awaitable,
     Callable,
+    Coroutine,
     Dict,
     Iterable,
     List,
     Optional,
     Sequence,
-    TYPE_CHECKING,
     Tuple,
     Type,
     TypeVar,
     Union,
 )
 
-import boto3
 import aiobotocore.session
-
+import boto3
 from meadowrun.aws_integration.aws_core import _get_default_region_name
 from meadowrun.aws_integration.aws_permissions_install import _EC2_ROLE_INSTANCE_PROFILE
 from meadowrun.aws_integration.ec2 import (
@@ -40,10 +41,9 @@ from meadowrun.aws_integration.ec2_ssh_keys import (
     get_meadowrun_ssh_key,
 )
 from meadowrun.aws_integration.grid_tasks_sqs import (
-    _add_worker_shutdown_messages,
-    _add_tasks,
-    _create_request_queue,
-    receive_results,
+    add_tasks,
+    add_worker_shutdown_messages,
+    create_request_queue,
     retry_task,
     worker_function,
 )
@@ -62,14 +62,14 @@ from meadowrun.aws_integration.management_lambdas.ec2_alloc_stub import (
     _RUNNING_JOBS,
     ignore_boto3_error_code,
 )
+from meadowrun.aws_integration.s3 import S3ObjectStorage, _get_bucket_name
 from meadowrun.instance_allocation import (
     InstanceRegistrar,
     _InstanceState,
     _TInstanceState,
     allocate_single_job_to_instance,
 )
-from meadowrun.instance_selection import ResourcesInternal, CloudInstance
-from meadowrun.aws_integration.s3 import S3ObjectStorage
+from meadowrun.instance_selection import CloudInstance, ResourcesInternal
 from meadowrun.run_job_core import (
     AllocVM,
     CloudProviderType,
@@ -79,13 +79,16 @@ from meadowrun.run_job_core import (
     SshHost,
     WaitOption,
 )
+from meadowrun.s3_grid_job import receive_results
 
 if TYPE_CHECKING:
     from types import TracebackType
-    from typing_extensions import Literal
+
     from meadowrun.meadowrun_pb2 import Job, ProcessState
+    from meadowrun.run_job_local import AgentTaskWorkerServer, Stats
     from types_aiobotocore_s3 import S3Client
     from types_aiobotocore_sqs import SQSClient
+    from typing_extensions import Literal
 
 # SEE ALSO ec2_alloc_stub.py
 
@@ -96,26 +99,26 @@ _U = TypeVar("_U")
 # replicate into each region.
 _AMIS = {
     "plain": {
-        "us-east-2": "ami-0c2ec58a802340753",
-        "us-east-1": "ami-0a519867d2df1d0db",
-        "us-west-1": "ami-036f320911e0afa56",
-        "us-west-2": "ami-09a6c93dc39a05f37",
-        "eu-central-1": "ami-003339e724b9bfdfd",
-        "eu-west-1": "ami-085fa49ba176751b4",
-        "eu-west-2": "ami-08345d3288b4f44c7",
-        "eu-west-3": "ami-07824146798815dfd",
-        "eu-north-1": "ami-09769400e1855038c",
+        "us-east-2": "ami-07f0f36fa29c65aaa",
+        "us-east-1": "ami-0ee6b624d61b9b641",
+        "us-west-1": "ami-06f0dd65aff3c563b",
+        "us-west-2": "ami-02d0917b63f6b9bae",
+        "eu-central-1": "ami-0e77a4d360d7b7610",
+        "eu-west-1": "ami-040289dd65de37dda",
+        "eu-west-2": "ami-03c5174fcd071728e",
+        "eu-west-3": "ami-0af288ea3afdeb8a6",
+        "eu-north-1": "ami-0cdd55e01ba94dbdd",
     },
     "cuda": {
-        "us-east-2": "ami-05fc9761c08a10e8d",
-        "us-east-1": "ami-07fb980d159f20b8a",
-        "us-west-1": "ami-0a45b65eefcc590e3",
-        "us-west-2": "ami-0762674ff454dc3a7",
-        "eu-central-1": "ami-03d7b166252674d75",
-        "eu-west-1": "ami-0a39a88acace4f75f",
-        "eu-west-2": "ami-042bfbb3a9039d3ac",
-        "eu-west-3": "ami-02500f90feb40d2e2",
-        "eu-north-1": "ami-0b0f26cf57f54e286",
+        "us-east-2": "ami-0ab0964aab023b6b2",
+        "us-east-1": "ami-0e9c845879ca81e49",
+        "us-west-1": "ami-00cbbc06817d02c38",
+        "us-west-2": "ami-00e359519a1e690e6",
+        "eu-central-1": "ami-06ec989f961db4c5c",
+        "eu-west-1": "ami-0bd8d3622c2af4e82",
+        "eu-west-2": "ami-0b88f83ef01a49dd8",
+        "eu-west-3": "ami-061cbd409324c685d",
+        "eu-north-1": "ami-030f6e8762e7b6220",
     },
 }
 SSH_USER = "ubuntu"
@@ -719,7 +722,11 @@ class EC2GridJobInterface(GridJobCloudInterface):
 
         self._ssh_private_key = get_meadowrun_ssh_key(self._region_name)
 
-        self._request_queue_url: Optional[asyncio.Task[str]] = None
+        # We keep multiple request queues so that we can retry tasks with more
+        # resources. The ith queue (where i starts at 0) will have workers with
+        # (original_memory * (i + 1)) memory. For the happy path, we will only create
+        # one queue with one set of workers having the originally requested memory
+        self._request_queue_urls: List[asyncio.Task[str]] = []
         self._task_argument_ranges: Optional[asyncio.Task[List[Tuple[int, int]]]] = None
 
         # this id is just used for creating the job's queues. It has no relationship to
@@ -753,19 +760,36 @@ class EC2GridJobInterface(GridJobCloudInterface):
     def create_instance_registrar(self) -> InstanceRegistrar:
         return EC2InstanceRegistrar(self._region_name, "create")
 
+    def create_queue(self) -> int:
+        if self._sqs_client is None:
+            raise ValueError("EC2GridJobInterface must be created with `async with`")
+
+        index = len(self._request_queue_urls)
+        if index == 0:
+            job_id_modified = self._job_id
+        else:
+            job_id_modified = f"{self._job_id}-{index}"
+        self._request_queue_urls.append(
+            asyncio.create_task(create_request_queue(job_id_modified, self._sqs_client))
+        )
+        return index
+
     async def setup_and_add_tasks(self, tasks: Sequence[_T]) -> None:
         if self._sqs_client is None or self._s3_client is None:
             raise ValueError("EC2GridJobInterface must be created with `async with`")
 
         # create SQS queues and add tasks to the request queue
         print(f"The current run_map's id is {self._job_id}")
-        self._request_queue_url = asyncio.create_task(
-            _create_request_queue(self._job_id, self._sqs_client)
-        )
+        queue_index = self.create_queue()
+        if queue_index != 0:
+            raise ValueError(
+                "setup_and_add_tasks was called more than once, or create_queue was "
+                "called before setup_and_add_tasks"
+            )
         self._task_argument_ranges = asyncio.create_task(
-            _add_tasks(
+            add_tasks(
                 self._job_id,
-                await self._request_queue_url,
+                await self._request_queue_urls[queue_index],
                 self._s3_client,
                 self._sqs_client,
                 tasks,
@@ -782,29 +806,27 @@ class EC2GridJobInterface(GridJobCloudInterface):
             instance_name,
         )
 
-    async def shutdown_workers(self, num_workers: int) -> None:
-        if self._request_queue_url is None:
-            raise ValueError(
-                "setup_and_add_tasks must be called before send_worker_done_messages"
-            )
+    async def shutdown_workers(self, num_workers: int, queue_index: int) -> None:
+        if len(self._request_queue_urls) < queue_index + 1:
+            raise ValueError(f"Queue {queue_index} has not been created yet")
         if self._sqs_client is None:
             raise ValueError("EC2GridJobInterface must be created with `async with`")
-        await _add_worker_shutdown_messages(
-            await self._request_queue_url, num_workers, self._sqs_client
+        await add_worker_shutdown_messages(
+            await self._request_queue_urls[queue_index], num_workers, self._sqs_client
         )
 
     async def get_worker_function(
-        self, user_function: Callable[[_T], _U]
-    ) -> Callable[[str, str], None]:
-        if self._request_queue_url is None:
-            raise ValueError(
-                "setup_and_add_tasks must be called before get_worker_function"
-            )
+        self, queue_index: int
+    ) -> Callable[
+        [str, str, AgentTaskWorkerServer, Callable[[], Awaitable[Stats]]],
+        Coroutine[Any, Any, None],
+    ]:
+        if len(self._request_queue_urls) < queue_index + 1:
+            raise ValueError(f"Queue {queue_index} has not been created yet")
 
         return functools.partial(
             worker_function,
-            user_function,
-            await self._request_queue_url,
+            await self._request_queue_urls[queue_index],
             self._job_id,
             self._region_name,
         )
@@ -815,22 +837,40 @@ class EC2GridJobInterface(GridJobCloudInterface):
         if self._s3_client is None:
             raise ValueError("EC2GridJobInterface must be created with `async with`")
 
+        # Note: download here is all via S3. It's important that downloads are fast
+        # enough - in some cases (many workers, small-ish tasks) the rate of downloading
+        # the results to the client can be a limiting factor. Other alternatives that
+        # were considered: - SQS queues. Limitations are 1. Only 10 messages can be
+        # downloaded at a time, which slows us down when we have hundreds of tasks
+        # completing quickly. 2. SQS message limit is 256KB which means we need to use
+        # S3 to transfer the actual data in the case of large results. - Sending data
+        # via SSH/SCP. This seems like it should be faster especially in the same
+        # VPC/subnet, but even in that case, uploading to S3 first seems to be faster.
+        # It's possible this would make sense in cases where we e.g. overwhelm the S3
+        # bucket.
+
         return receive_results(
-            self._job_id,
-            self._region_name,
             self._s3_client,
+            _get_bucket_name(self._region_name),
+            self._job_id,
             stop_receiving=stop_receiving,
             all_workers_exited=workers_done,
+            # it's rare for any results to be ready in <4 seconds
+            initial_wait_seconds=4,
         )
 
-    async def retry_task(self, task_id: int, attempts_so_far: int) -> None:
-        if self._request_queue_url is None or self._task_argument_ranges is None:
+    async def retry_task(
+        self, task_id: int, attempts_so_far: int, queue_index: int
+    ) -> None:
+        if len(self._request_queue_urls) < queue_index + 1:
+            raise ValueError(f"Queue {queue_index} has not been created yet")
+        if self._task_argument_ranges is None:
             raise ValueError("setup_and_add_tasks must be called before retry_task")
         if self._sqs_client is None:
             raise ValueError("EC2GridJobInterface must be created with `async with`")
 
         await retry_task(
-            await self._request_queue_url,
+            await self._request_queue_urls[queue_index],
             task_id,
             attempts_so_far + 1,
             (await self._task_argument_ranges)[task_id],
