@@ -9,7 +9,9 @@ import uuid
 from typing import (
     Any,
     AsyncIterable,
+    Awaitable,
     Callable,
+    Coroutine,
     Generic,
     Iterable,
     List,
@@ -30,12 +32,12 @@ from meadowrun.azure_integration.azure_ssh_keys import ensure_meadowrun_key_pair
 from meadowrun.azure_integration.blob_storage import AzureBlobStorage
 from meadowrun.azure_integration.grid_tasks_queue import (
     Queue,
-    _add_tasks,
-    _add_worker_shutdown_message,
-    _create_queues_for_job,
-    _retry_task,
+    add_tasks,
+    add_worker_shutdown_message,
+    create_queues_for_job,
+    retry_task,
     get_results_unordered,
-    worker_function,
+    worker_function_async,
 )
 from meadowrun.azure_integration.mgmt_functions.azure_constants import (
     ALLOCATED_TIME,
@@ -82,6 +84,7 @@ if TYPE_CHECKING:
     from typing_extensions import Literal
     from types import TracebackType
     from meadowrun.meadowrun_pb2 import Job, ProcessState
+    from meadowrun.run_job_local import AgentTaskWorkerServer, Stats
 
 
 _T = TypeVar("_T")
@@ -531,6 +534,11 @@ class AzureVMGridJobInterface(GridJobCloudInterface, Generic[_T, _U]):
     issues.
     """
 
+    def create_queue(self) -> int:
+        raise NotImplementedError(
+            "Retrying with more resources is not supported on Azure yet"
+        )
+
     def __init__(self, alloc_cloud_instance: AllocAzureVM):
         self._cloud_provider = alloc_cloud_instance.get_cloud_provider()
         self._location = alloc_cloud_instance._get_location()
@@ -556,10 +564,10 @@ class AzureVMGridJobInterface(GridJobCloudInterface, Generic[_T, _U]):
             ensure_meadowrun_key_pair(self._location)
         )
         self._request_result_queues = asyncio.create_task(
-            _create_queues_for_job(self._job_id, self._location)
+            create_queues_for_job(self._job_id, self._location)
         )
         self._tasks = tasks
-        await _add_tasks((await self._request_result_queues)[0], tasks)
+        await add_tasks((await self._request_result_queues)[0], tasks)
 
     async def ssh_host_from_address(self, address: str, instance_name: str) -> SshHost:
         if self._ssh_private_key is None:
@@ -575,29 +583,41 @@ class AzureVMGridJobInterface(GridJobCloudInterface, Generic[_T, _U]):
             instance_name,
         )
 
-    async def shutdown_workers(self, num_workers: int) -> None:
+    async def shutdown_workers(self, num_workers: int, queue_index: int) -> None:
+        if queue_index != 0:
+            raise NotImplementedError(
+                "Retrying with more resources is not implemented for Azure yet"
+            )
+
         if self._request_result_queues is None:
             raise ValueError(
                 "Must call setup_and_add_tasks before calling shutdown_workers"
             )
 
-        await _add_worker_shutdown_message(
+        await add_worker_shutdown_message(
             (await self._request_result_queues)[0], num_workers
         )
 
     async def get_worker_function(
-        self, user_function: Callable[[_T], _U]
-    ) -> Callable[[str, str], None]:
+        self, queue_index: int
+    ) -> Callable[
+        [str, str, AgentTaskWorkerServer, Callable[[], Awaitable[Stats]]],
+        Coroutine[Any, Any, None],
+    ]:
         if self._request_result_queues is None:
             raise ValueError(
                 "Must call setup_and_add_tasks before calling get_worker_function"
             )
+        if queue_index != 0:
+            raise NotImplementedError(
+                "Retrying with more resources is not implemented for Azure yet"
+            )
 
         request_queue, result_queue = await self._request_result_queues
 
         return functools.partial(
-            worker_function,
-            user_function,
+            worker_function_async,
+            # user_function,
             request_queue,
             result_queue,
         )
@@ -617,11 +637,13 @@ class AzureVMGridJobInterface(GridJobCloudInterface, Generic[_T, _U]):
             workers_done,
         )
 
-    async def retry_task(self, task_id: int, attempts_so_far: int) -> None:
+    async def retry_task(
+        self, task_id: int, attempts_so_far: int, queue_index: int
+    ) -> None:
         if self._tasks is None or self._request_result_queues is None:
             raise ValueError("Must call setup_and_add_tasks before calling retry_task")
 
-        await _retry_task(
+        await retry_task(
             (await self._request_result_queues)[0],
             task_id,
             attempts_so_far + 1,
