@@ -38,7 +38,6 @@ from meadowrun.docker_controller import (
     pull_image,
     push_image,
 )
-from meadowrun.func_worker_storage_helper import FuncWorkerClientObjectStorage
 from meadowrun.meadowrun_pb2 import (
     CodeZipFile,
     EnvironmentSpec,
@@ -57,10 +56,11 @@ from meadowrun.run_job_core import (
     ContainerRegistryHelper,
     LocalObjectStorage,
 )
-from meadowrun.func_worker_storage_helper import (
+from meadowrun.s3_grid_job import (
     try_get_storage_file,
     write_storage_file,
 )
+from meadowrun.func_worker_storage_helper import FuncWorkerClientObjectStorage
 import meadowrun.func_worker_storage_helper
 
 _GIT_REPO_URL_SUFFIXES_TO_REMOVE = [".git", "/"]
@@ -280,10 +280,10 @@ async def _get_zip_file_code_paths(
     if decoded_url.scheme not in _ALL_OBJECT_STORAGES:
         raise ValueError(f"Unknown URL scheme in {code_zip_file.url}")
 
-    object_storage = _ALL_OBJECT_STORAGES[decoded_url.scheme]()
-    extracted_folder = await object_storage.download_and_unzip(
-        code_zip_file.url, local_copies_folder
-    )
+    async with _ALL_OBJECT_STORAGES[decoded_url.scheme]() as object_storage:
+        extracted_folder = await object_storage.download_and_unzip(
+            code_zip_file.url, local_copies_folder
+        )
 
     return (
         [
@@ -464,7 +464,13 @@ async def compile_environment_spec_locally(
 
     if environment_spec.environment_type == EnvironmentType.CONDA:
         get_cached_or_create: Callable[
-            [str, str, str, Callable[[str, str], bool], Callable[[str, str], None]],
+            [
+                str,
+                str,
+                str,
+                Callable[[str, str], Awaitable[bool]],
+                Callable[[str, str], Awaitable[None]],
+            ],
             Awaitable[str],
         ] = get_cached_or_create_conda_environment
     elif environment_spec.environment_type == EnvironmentType.PIP:
