@@ -3,17 +3,20 @@ from __future__ import annotations
 import asyncio
 import asyncio.subprocess
 import dataclasses
+import functools
 import itertools
 import os
 import os.path
 import pathlib
 import pickle
 import shutil
+import struct
 import sys
 import traceback
 from typing import (
     Any,
     AsyncIterable,
+    Awaitable,
     Callable,
     Coroutine,
     Dict,
@@ -24,14 +27,19 @@ from typing import (
     TYPE_CHECKING,
     Tuple,
     TypeVar,
+    Union,
 )
 
+import psutil
+
 from meadowrun.aws_integration.ecr import get_ecr_username_password
 from meadowrun.azure_integration.acr import get_acr_username_password
 
 if TYPE_CHECKING:
     from typing_extensions import Literal
 
+    JobSpecType = Literal["py_command", "py_function", "py_agent"]
+
     from meadowrun._vendor import aiodocker
     from meadowrun._vendor.aiodocker import containers as aiodocker_containers
     from meadowrun.instance_selection import ResourcesInternal
@@ -65,6 +73,7 @@ from meadowrun.meadowrun_pb2 import (
     Job,
     ProcessState,
     PyFunctionJob,
+    PyAgentJob,
     StringPair,
 )
 from meadowrun.run_job_core import (
@@ -120,6 +129,7 @@ class _JobSpecTransformed:
     environment_variables: Dict[str, str] = dataclasses.field(
         default_factory=lambda: {}
     )
+    server: Optional[AgentTaskWorkerServer] = None
 
 
 def _io_file_container_binds(
@@ -200,7 +210,7 @@ _FUNC_WORKER_PATH = str(
 
 
 def _prepare_function(
-    job_id: str, function: PyFunctionJob, io_folder: str
+    job_id: str, function: Union[PyFunctionJob, PyAgentJob], io_folder: str
 ) -> Tuple[Sequence[str], Sequence[str]]:
     """
     Creates files in io_folder for the child process to use and returns (command line
@@ -304,8 +314,81 @@ def _prepare_py_function(
     )
 
 
+_TASK_WORKER_PATH = str(
+    (
+        pathlib.Path(__file__).parent / "func_worker" / "__meadowrun_task_worker.py"
+    ).resolve()
+)
+
+
+async def _prepare_py_agent(
+    job: Job, io_folder: str, is_container: bool
+) -> _JobSpecTransformed:
+    """
+    Creates files in io_folder for the child process to use and returns
+    _JobSpecTransformed. We use __meadowrun_task_worker to start the function in the
+    child process.
+    """
+
+    io_files = [
+        # these line up with __meadowrun_task_worker
+        os.path.join(job.job_id + ".state"),
+        os.path.join(job.job_id + ".result"),
+    ]
+
+    if not is_container:
+        worker_path = _TASK_WORKER_PATH
+        io_path_container = os.path.join(io_folder, job.job_id)
+    else:
+        worker_path = (
+            f"{MEADOWRUN_CODE_MOUNT_LINUX}{os.path.basename(_TASK_WORKER_PATH)}"
+        )
+        io_path_container = f"{MEADOWRUN_IO_MOUNT_LINUX}/{job.job_id}"
+        for io_file in io_files:
+            open(os.path.join(io_folder, io_file), "w", encoding="utf-8").close()
+
+    server = await agent_start_serving("0.0.0.0" if is_container else "127.0.0.1")
+    command_line = [
+        "python",
+        worker_path,
+        "--result-highest-pickle-protocol",
+        str(job.result_highest_pickle_protocol),
+        "--io-path",
+        io_path_container,
+        "--host",
+        "host.docker.internal" if is_container else "127.0.0.1",
+        "--port",
+        str(server.port),
+    ]
+
+    command_line_for_function, io_files_for_function = _prepare_function(
+        job.job_id, job.py_agent, io_folder
+    )
+
+    return _JobSpecTransformed(
+        list(itertools.chain(command_line, command_line_for_function)),
+        _io_file_container_binds(
+            io_folder,
+            itertools.chain(io_files, io_files_for_function),
+        )
+        + [(_TASK_WORKER_PATH, worker_path)],
+        server=server,
+    )
+
+
+@dataclasses.dataclass
+class StatsAccumulator:
+    max_memory_used_gb: float = 0
+
+
+@dataclasses.dataclass(frozen=True)
+class Stats:
+    # TODO add cpu usage
+    memory_in_use_gb: float
+
+
 async def _launch_non_container_job(
-    job_spec_type: Literal["py_command", "py_function"],
+    job_spec_type: JobSpecType,
     job_spec_transformed: _JobSpecTransformed,
     code_paths: Sequence[str],
     cwd_path: Optional[str],
@@ -402,19 +485,35 @@ async def _launch_non_container_job(
     return process.pid, _non_container_job_continuation(
         process,
         job_spec_type,
-        job.job_id,
+        job_spec_transformed,
+        job,
         io_folder,
-        job.result_highest_pickle_protocol,
         log_file_name,
     )
 
 
+async def _stop_process(process: asyncio.subprocess.Process) -> None:
+    """Safe way to terminate and kill a process."""
+    try:
+        process.terminate()
+        process.kill()
+    except ProcessLookupError:
+        # if it's already gone, we're done here
+        pass
+
+
+async def _get_non_container_stats(pid: int) -> Stats:
+    process = psutil.Process(pid)
+    # RSS is in bytes: https://man7.org/linux/man-pages/man1/ps.1.html
+    return Stats(process.memory_info().rss / (1024**3))
+
+
 async def _non_container_job_continuation(
     process: asyncio.subprocess.Process,
-    job_spec_type: Literal["py_command", "py_function"],
-    job_id: str,
+    job_spec_type: JobSpecType,
+    job_spec_transformed: _JobSpecTransformed,
+    job: Job,
     io_folder: str,
-    result_highest_pickle_protocol: int,
     log_file_name: str,
 ) -> ProcessState:
     """
@@ -424,15 +523,26 @@ async def _non_container_job_continuation(
     """
 
     try:
-        # wait for the process to finish
-        # TODO add an optional timeout
 
-        async for line in process.stdout:  # type: ignore
-            sys.stdout.buffer.write(line)
+        async def tail_log() -> None:
+            async for line in process.stdout:  # type: ignore
+                print(line)
+
+        tail_log_task = asyncio.create_task(tail_log())
+
+        if job_spec_type == "py_agent":
+            await _run_agent(
+                job_spec_transformed,
+                job,
+                functools.partial(_get_non_container_stats, process.pid),
+            )
+            await _stop_process(process)
+
         returncode = await process.wait()
+        await tail_log_task
         return _completed_job_state(
             job_spec_type,
-            job_id,
+            job.job_id,
             io_folder,
             log_file_name,
             returncode,
@@ -442,15 +552,18 @@ async def _non_container_job_continuation(
     except asyncio.CancelledError:
         raise
     except Exception as e:
-        # there was an exception while trying to get the final ProcessState
+        print(
+            "There was an exception while trying to get the final ProcessState:\n"
+            + traceback.format_exc()
+        )
         return ProcessState(
             state=ProcessStateEnum.ERROR_GETTING_STATE,
-            pickled_result=pickle_exception(e, result_highest_pickle_protocol),
+            pickled_result=pickle_exception(e, job.result_highest_pickle_protocol),
         )
 
 
 async def _launch_container_job(
-    job_spec_type: Literal["py_command", "py_function"],
+    job_spec_type: JobSpecType,
     container_image_name: str,
     job_spec_transformed: _JobSpecTransformed,
     code_paths: Sequence[str],
@@ -588,21 +701,39 @@ async def _launch_container_job(
         container,
         docker_client,
         job_spec_type,
-        job.job_id,
+        job_spec_transformed,
+        job,
+        # job.job_id,
         io_folder,
-        job.result_highest_pickle_protocol,
+        # job.result_highest_pickle_protocol,
         log_file_name,
         sidecar_containers,
     )
 
 
+async def _get_container_stats(
+    container: aiodocker_containers.DockerContainer,
+) -> Stats:
+    stats = await container.stats(stream=False)
+    if not stats:
+        return Stats(0)
+    # according to
+    # https://docs.docker.com/engine/api/v1.41/#tag/Container/operation/ContainerStats
+    # "memory used" in the docker CLI corresponds to this formula:
+    return Stats(
+        (stats[0]["memory_stats"]["usage"] - stats[0]["memory_stats"]["stats"]["cache"])
+        / (1024**3)
+    )
+
+
 async def _container_job_continuation(
     container: aiodocker_containers.DockerContainer,
     docker_client: aiodocker.Docker,
-    job_spec_type: Literal["py_command", "py_function"],
-    job_id: str,
+    job_spec_type: JobSpecType,
+    job_spec_transformed: _JobSpecTransformed,
+    job: Job,
     io_folder: str,
-    result_highest_pickle_protocol: int,
+    # result_highest_pickle_protocol: int,
     log_file_name: str,
     sidecar_containers: List[aiodocker_containers.DockerContainer],
 ) -> ProcessState:
@@ -614,22 +745,37 @@ async def _container_job_continuation(
     docker_client just needs to be closed when the container process has completed.
     """
     try:
-        # Docker appears to have an objection to having a log driver that can produce
-        # plain text files (https://github.com/moby/moby/issues/17020) so we implement
-        # that in a hacky way here.
-        # TODO figure out overall strategy for logging, maybe eventually implement our
-        #  own plain text/whatever log driver for docker.
-        async for line in container.log(stdout=True, stderr=True, follow=True):
-            print(line, end="")
+
+        async def tail_log() -> None:
+            # 1. Docker appears to have an objection to having a log driver that can
+            # produce plain text files (https://github.com/moby/moby/issues/17020) so we
+            # implement that in a hacky way here.
+            # TODO figure out overall strategy for
+            # logging, maybe eventually implement our own plain text/whatever log driver
+            #  for docker.
+            async for line in container.log(stdout=True, stderr=True, follow=True):
+                print(f"Task worker: {line}", end="")
+
+        tail_log_task = asyncio.create_task(tail_log())
+
+        if job_spec_type == "py_agent":
+            await _run_agent(
+                job_spec_transformed,
+                job,
+                functools.partial(_get_container_stats, container),
+            )
+            await container.stop()
 
         wait_result = await container.wait()
+        await tail_log_task
+
         # as per https://docs.docker.com/engine/api/v1.41/#operation/ContainerWait we
         # can get the return code from the result
         return_code = wait_result["StatusCode"]
 
         return _completed_job_state(
             job_spec_type,
-            job_id,
+            job.job_id,
             io_folder,
             log_file_name,
             return_code,
@@ -639,10 +785,13 @@ async def _container_job_continuation(
     except asyncio.CancelledError:
         raise
     except Exception as e:
-        # there was an exception while trying to get the final ProcessState
+        print(
+            "There was an exception while trying to get the final ProcessState:\n"
+            + traceback.format_exc()
+        )
         return ProcessState(
             state=ProcessStateEnum.ERROR_GETTING_STATE,
-            pickled_result=pickle_exception(e, result_highest_pickle_protocol),
+            pickled_result=pickle_exception(e, job.result_highest_pickle_protocol),
         )
     finally:
         try:
@@ -655,8 +804,29 @@ async def _container_job_continuation(
         await docker_client.__aexit__(None, None, None)
 
 
+async def _run_agent(
+    job_spec_transformed: _JobSpecTransformed,
+    job: Job,
+    get_worker_stats: Callable[[], Awaitable[Stats]],
+) -> None:
+    # run the agent function. The agent function connects to another
+    # process, the task worker, which runs the actual user function.
+    assert job_spec_transformed.server is not None
+    agent_func = pickle.loads(job.py_agent.pickled_agent_function)
+    agent_func_args, agent_func_kwargs = pickle.loads(
+        job.py_agent.pickled_agent_function_arguments
+    )
+    await agent_func(
+        *agent_func_args,
+        worker_server=job_spec_transformed.server,
+        get_worker_stats=get_worker_stats,
+        **agent_func_kwargs,
+    )
+    await job_spec_transformed.server.close()
+
+
 def _completed_job_state(
-    job_spec_type: Literal["py_command", "py_function"],
+    job_spec_type: JobSpecType,
     job_id: str,
     io_folder: str,
     log_file_name: str,
@@ -682,8 +852,8 @@ def _completed_job_state(
 
     # if we returned normally
 
-    # for py_funcs, get the state
-    if job_spec_type == "py_function":
+    # for py_funcs or py_agents, get the state
+    if job_spec_type in ["py_function", "py_agent"]:
         state_file = os.path.join(io_folder, job_id + ".state")
         with open(state_file, "r", encoding="utf-8") as state_file_text_reader:
             state_string = state_file_text_reader.read()
@@ -721,28 +891,14 @@ def _completed_job_state(
     )
 
 
-def _get_default_working_folder() -> str:
-    # figure out the default working_folder based on the OS
-    if os.name == "nt":
-        return os.path.join(os.environ["USERPROFILE"], "meadowrun")
-    elif os.name == "posix":
-        return os.path.join(os.environ["HOME"], "meadowrun")
-    else:
-        raise ValueError(f"Unexpected os.name {os.name}")
-
-
-def _set_up_working_folder(
-    working_folder: Optional[str],
-) -> Tuple[str, str, str, str, str]:
+def _set_up_working_folder() -> Tuple[str, str, str, str, str]:
     """
     Sets the working_folder to a default if it's not set, creates the necessary
     subfolders, gets a machine-wide lock on the working folder, then returns io_folder,
     job_logs_folder, git_repos_folder, local_copies_folder
     """
 
-    if not working_folder:
-        working_folder = _get_default_working_folder()
-        os.makedirs(working_folder, exist_ok=True)
+    working_folder = "/var/meadowrun"
 
     # first, create the directories that we need
 
@@ -758,12 +914,6 @@ def _set_up_working_folder(
     # misc folder for e.g. storing environment export files sent from local machine
     misc_folder = os.path.join(working_folder, "misc")
 
-    os.makedirs(io_folder, exist_ok=True)
-    os.makedirs(job_logs_folder, exist_ok=True)
-    os.makedirs(git_repos_folder, exist_ok=True)
-    os.makedirs(local_copies_folder, exist_ok=True)
-    os.makedirs(misc_folder, exist_ok=True)
-
     return (
         io_folder,
         job_logs_folder,
@@ -909,9 +1059,74 @@ async def _get_credentials_for_job(
     )
 
 
+class AgentTaskWorkerServer:
+    def __init__(self) -> None:
+        self.server: Optional[asyncio.AbstractServer] = None
+        self.reader: Optional[asyncio.StreamReader] = None
+        self.writer: Optional[asyncio.StreamWriter] = None
+        self.have_connection: asyncio.Event = asyncio.Event()
+        self.port: Optional[int] = None
+
+    async def start_serving(
+        self,
+        host: str,
+    ) -> None:
+        server = await asyncio.start_server(self._handle_connection, host, 0)
+        (socket,) = server.sockets
+        self.port = socket.getsockname()[1]
+
+    async def _handle_connection(
+        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
+    ) -> None:
+        self.reader = reader
+        self.writer = writer
+        self.have_connection.set()
+
+    async def wait_for_task_worker_connection(
+        self, timeout: float = 30
+    ) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
+        await asyncio.wait_for(self.have_connection.wait(), timeout=timeout)
+        assert self.reader is not None
+        assert self.writer is not None
+        return self.reader, self.writer
+
+    async def send_message(self, bs: bytes) -> None:
+        if not self.have_connection.is_set():
+            await self.wait_for_task_worker_connection()
+        assert self.writer is not None
+        msg_len = struct.pack(">i", len(bs))
+        self.writer.write(msg_len)
+        self.writer.write(bs)
+        await self.writer.drain()
+
+    async def receive_message(self) -> Any:
+        if not self.have_connection.is_set():
+            await self.wait_for_task_worker_connection()
+        assert self.reader is not None
+        (result_len,) = struct.unpack(">i", await self.reader.read(4))
+        result_bs = await self.reader.read(result_len)
+        return pickle.loads(result_bs)
+
+    async def close(self) -> None:
+        if self.writer:
+            self.writer.close()
+            await self.writer.wait_closed()
+            self.have_connection.clear()
+        if self.server:
+            self.server.close()
+            await self.server.wait_closed()
+
+
+async def agent_start_serving(
+    host: str,
+) -> AgentTaskWorkerServer:
+    server = AgentTaskWorkerServer()
+    await server.start_serving(host)
+    return server
+
+
 async def run_local(
     job: Job,
-    working_folder: Optional[str] = None,
     cloud: Optional[Tuple[CloudProviderType, str]] = None,
     compile_environment_in_container: bool = True,
 ) -> Tuple[ProcessState, Optional[asyncio.Task[ProcessState]]]:
@@ -945,7 +1160,7 @@ async def run_local(
         git_repos_folder,
         local_copies_folder,
         misc_folder,
-    ) = _set_up_working_folder(working_folder)
+    ) = _set_up_working_folder()
 
     # the logging actually happens via stdout redirection in the run_job_local_main
     # caller
@@ -1026,6 +1241,8 @@ async def run_local(
             job_spec_transformed = _prepare_py_command(job, io_folder, is_container)
         elif job_spec_type == "py_function":
             job_spec_transformed = _prepare_py_function(job, io_folder, is_container)
+        elif job_spec_type == "py_agent":
+            job_spec_transformed = await _prepare_py_agent(job, io_folder, is_container)
         else:
             raise ValueError(f"Unknown job_spec {job_spec_type}")
 
@@ -1227,6 +1444,7 @@ class LocalHost(Host):
         pickle_protocol: int,
         wait_for_result: WaitOption,
         max_num_tasks_attempts: int,
+        retry_with_more_memory: bool,
     ) -> Optional[Sequence[_U]]:
         raise NotImplementedError("run_map on LocalHost is not implemented")
 
@@ -1240,6 +1458,7 @@ class LocalHost(Host):
         pickle_protocol: int,
         wait_for_result: WaitOption,
         max_num_tasks_attempts: int,
+        retry_with_more_memory: bool,
     ) -> AsyncIterable[TaskResult[_U]]:
         raise NotImplementedError(
             "run_map_as_completed is not implemented for LocalHost"
@@ -1247,3 +1466,20 @@ class LocalHost(Host):
 
     async def get_object_storage(self) -> ObjectStorage:
         return LocalObjectStorage()
+
+
+async def _accumulate_stats(
+    get_worker_stats: Callable[[], Awaitable[Stats]], accumulator: StatsAccumulator
+) -> None:
+    """Modifies accumulator in place!"""
+    while True:
+        try:
+            stats = await get_worker_stats()
+            if stats.memory_in_use_gb > accumulator.max_memory_used_gb:
+                accumulator.max_memory_used_gb = stats.memory_in_use_gb
+        except asyncio.CancelledError:
+            raise
+        except Exception:
+            print("Error updating stats:\n" + traceback.format_exc())
+
+        await asyncio.sleep(1)
